@@ -5,13 +5,13 @@
 
     The fastest markdown parser in pure Python with renderer feature.
 
-    :copyright: (c) 2014 by Hsiaoming Yang.
+    :copyright: (c) 2014 - 2015 by Hsiaoming Yang.
 """
 
 import re
 import inspect
 
-__version__ = '0.4.1'
+__version__ = '0.6'
 __author__ = 'Hsiaoming Yang <me@lepture.com>'
 __all__ = [
     'BlockGrammar', 'BlockLexer',
@@ -21,6 +21,17 @@ __all__ = [
 ]
 
 
+_key_pattern = re.compile(r'\s+')
+_escape_pattern = re.compile(r'&(?!#?\w+;)')
+_newline_pattern = re.compile(r'\r\n|\r')
+_inline_tag = (
+    r'a|em|strong|small|s|cite|q|dfn|abbr|data|time|code|'
+    r'var|samp|kbd|sub|sup|i|b|u|mark|ruby|rt|rp|bdi|bdo|'
+    r'span|br|wbr|ins|del|img|font'
+)
+_block_tag = r'(?!(?:%s)\b)\w+(?!:/|[^\w\s@]*@)\b' % _inline_tag
+
+
 def _pure_pattern(regex):
     pattern = regex.pattern
     if pattern.startswith('^'):
@@ -28,16 +39,10 @@ def _pure_pattern(regex):
     return pattern
 
 
-_key_pattern = re.compile(r'\s+')
-
-
 def _keyify(key):
     return _key_pattern.sub(' ', key.lower())
 
 
-_escape_pattern = re.compile(r'&(?!#?\w+;)')
-
-
 def escape(text, quote=False, smart_amp=True):
     """Replace special characters "&", "<" and ">" to HTML-safe sequences.
 
@@ -60,7 +65,7 @@ def escape(text, quote=False, smart_amp=True):
 
 
 def preprocessing(text, tab=4):
-    text = re.sub(r'\r\n|\r', '\n', text)
+    text = _newline_pattern.sub('\n', text)
     text = text.replace('\t', ' ' * tab)
     text = text.replace('\u00a0', ' ')
     text = text.replace('\u2424', '\n')
@@ -71,12 +76,6 @@ def preprocessing(text, tab=4):
 class BlockGrammar(object):
     """Grammars for block level tokens."""
 
-    _tag = (
-        r'(?!(?:'
-        r'a|em|strong|small|s|cite|q|dfn|abbr|data|time|code|'
-        r'var|samp|kbd|sub|sup|i|b|u|mark|ruby|rt|rp|bdi|bdo|'
-        r'span|br|wbr|ins|del|img)\b)\w+(?!:/|[^\w\s@]*@)\b'
-    )
     def_links = re.compile(
         r'^ *\[([^^\]]+)\]: *'  # [key]:
         r'<?([^\s>]+)>?'  # <link> or link
@@ -96,7 +95,7 @@ class BlockGrammar(object):
         r'([\s\S]+?)\s*'
         r'\1 *(?:\n+|$)'  # ```
     )
-    hrule = re.compile(r'^(?: *[-*_]){3,} *(?:\n+|$)')
+    hrule = re.compile(r'^ {0,3}[-*_](?: *[-*_]){2,} *(?:\n+|$)')
     heading = re.compile(r'^ *(#{1,6}) *([^\n]+?) *#* *(?:\n+|$)')
     lheading = re.compile(r'^([^\n]+)\n *(=|-)+ *(?:\n+|$)')
     block_quote = re.compile(r'^( *>[^\n]+(\n[^\n]+)*\n*)+')
@@ -133,14 +132,14 @@ class BlockGrammar(object):
             _pure_pattern(block_quote),
             _pure_pattern(def_links),
             _pure_pattern(def_footnotes),
-            '<' + _tag,
+            '<' + _block_tag,
         )
     )
     block_html = re.compile(
         r'^ *(?:%s|%s|%s) *(?:\n{2,}|\s*$)' % (
             r'<!--[\s\S]*?-->',
-            r'<(%s)[\s\S]+?<\/\1>' % _tag,
-            r'''<%s(?:"[^"]*"|'[^']*'|[^'">])*?>''' % _tag,
+            r'<(%s)[\s\S]+?<\/\1>' % _block_tag,
+            r'''<%s(?:"[^"]*"|'[^']*'|[^'">])*?>''' % _block_tag,
         )
     )
     table = re.compile(
@@ -154,20 +153,21 @@ class BlockGrammar(object):
 
 class BlockLexer(object):
     """Block level lexer for block grammars."""
+    grammar_class = BlockGrammar
 
-    default_features = [
-        'newline', 'block_code', 'fences', 'heading',
-        'nptable', 'lheading', 'hrule', 'block_quote',
+    default_rules = [
+        'newline', 'hrule', 'block_code', 'fences', 'heading',
+        'nptable', 'lheading', 'block_quote',
         'list_block', 'block_html', 'def_links',
         'def_footnotes', 'table', 'paragraph', 'text'
     ]
 
-    list_features = (
+    list_rules = (
         'newline', 'block_code', 'fences', 'lheading', 'hrule',
         'block_quote', 'list_block', 'block_html', 'text',
     )
 
-    footnote_features = (
+    footnote_rules = (
         'newline', 'block_code', 'fences', 'heading',
         'nptable', 'lheading', 'hrule', 'block_quote',
         'list_block', 'block_html', 'table', 'paragraph', 'text'
@@ -179,42 +179,42 @@ class BlockLexer(object):
         self.def_footnotes = {}
 
         if not rules:
-            rules = BlockGrammar()
+            rules = self.grammar_class()
 
         self.rules = rules
 
-    def __call__(self, text, features=None):
-        return self.parse(text, features)
+    def __call__(self, text, rules=None):
+        return self.parse(text, rules)
 
-    def parse(self, text, features=None):
+    def parse(self, text, rules=None):
         text = text.rstrip('\n')
 
-        if not features:
-            features = self.default_features
+        if not rules:
+            rules = self.default_rules
 
         def manipulate(text):
-            for key in features:
+            for key in rules:
                 rule = getattr(self.rules, key)
                 m = rule.match(text)
                 if not m:
                     continue
                 getattr(self, 'parse_%s' % key)(m)
                 return m
-            return False
+            return False  # pragma: no cover
 
         while text:
             m = manipulate(text)
             if m is not False:
                 text = text[len(m.group(0)):]
                 continue
-            if text:
+            if text:  # pragma: no cover
                 raise RuntimeError('Infinite loop at: %s' % text)
         return self.tokens
 
     def parse_newline(self, m):
         length = len(m.group(0))
         if length > 1:
-            self.tokens.append({'type': 'space'})
+            self.tokens.append({'type': 'newline'})
 
     def parse_block_code(self, m):
         code = m.group(0)
@@ -284,8 +284,10 @@ class BlockLexer(object):
             loose = _next
             if not loose and re.search(r'\n\n(?!\s*$)', item):
                 loose = True
-            if i != length - 1:
-                _next = item[len(item)-1] == '\n'
+
+            rest = len(item)
+            if i != length - 1 and rest:
+                _next = item[rest-1] == '\n'
                 if not loose:
                     loose = _next
 
@@ -296,7 +298,7 @@ class BlockLexer(object):
 
             self.tokens.append({'type': t})
             # recurse
-            self.parse(item, self.list_features)
+            self.parse(item, self.list_rules)
             self.tokens.append({'type': 'list_item_end'})
 
     def parse_block_quote(self, m):
@@ -341,7 +343,7 @@ class BlockLexer(object):
                 newlines.append(line[whitespace:])
             text = '\n'.join(newlines)
 
-        self.parse(text, self.footnote_features)
+        self.parse(text, self.footnote_rules)
 
         self.tokens.append({
             'type': 'footnote_end',
@@ -416,17 +418,19 @@ class InlineGrammar(object):
     """Grammars for inline level tokens."""
 
     escape = re.compile(r'^\\([\\`*{}\[\]()#+\-.!_>~|])')  # \* \+ \! ....
-    tag = re.compile(
-        r'^<!--[\s\S]*?-->|'  # comment
-        r'^<\/\w+>|'  # close tag
-        r'^<\w+[^>]*?>'  # open tag
+    inline_html = re.compile(
+        r'^(?:%s|%s|%s)' % (
+            r'<!--[\s\S]*?-->',
+            r'<(%s)[\s\S]+?<\/\1>' % _inline_tag,
+            r'''<(?:%s)(?:"[^"]*"|'[^']*'|[^'">])*?>''' % _inline_tag,
+        )
     )
-    autolink = re.compile(r'^<([^ >]+(@|:\/)[^ >]+)>')
+    autolink = re.compile(r'^<([^ >]+(@|:)[^ >]+)>')
     link = re.compile(
         r'^!?\[('
         r'(?:\[[^^\]]*\]|[^\[\]]|\](?=[^\[]*\]))*'
         r')\]\('
-        r'''\s*<?([\s\S]*?)>?(?:\s+['"]([\s\S]*?)['"])?\s*'''
+        r'''\s*(<)?([\s\S]*?)(?(2)>)(?:\s+['"]([\s\S]*?)['"])?\s*'''
         r'\)'
     )
     reflink = re.compile(
@@ -448,7 +452,7 @@ class InlineGrammar(object):
     )
     code = re.compile(r'^(`+)\s*([\s\S]*?[^`])\s*\1(?!`)')  # `code`
     linebreak = re.compile(r'^ {2,}\n(?!\s*$)')
-    strikethrough = re.compile(r'^~~(?=\S)([\s\S]*?\S)~~')  # ~~word~~
+    strikethrough = re.compile(r'^~~(?=\S)([\s\S]+?\S)~~')  # ~~word~~
     footnote = re.compile(r'^\[\^([^\]]+)\]')
     text = re.compile(r'^[\s\S]+?(?=[\\<!\[_*`~]|https?://| {2,}\n|$)')
 
@@ -464,9 +468,10 @@ class InlineGrammar(object):
 
 class InlineLexer(object):
     """Inline level lexer for inline grammars."""
+    grammar_class = InlineGrammar
 
-    default_features = [
-        'escape', 'autolink', 'url', 'tag',
+    default_rules = [
+        'escape', 'inline_html', 'autolink', 'url',
         'footnote', 'link', 'reflink', 'nolink',
         'double_emphasis', 'emphasis', 'code',
         'linebreak', 'strikethrough', 'text',
@@ -479,7 +484,7 @@ class InlineLexer(object):
         self.footnote_index = 0
 
         if not rules:
-            rules = InlineGrammar()
+            rules = self.grammar_class()
 
         self.rules = rules
 
@@ -494,18 +499,18 @@ class InlineLexer(object):
         self.links = links or {}
         self.footnotes = footnotes or {}
 
-    def output(self, text, features=None):
+    def output(self, text, rules=None):
         text = text.rstrip('\n')
-        if not features:
-            features = list(self.default_features)
+        if not rules:
+            rules = list(self.default_rules)
 
-        if self._in_footnote and 'footnote' in features:
-            features.remove('footnote')
+        if self._in_footnote and 'footnote' in rules:
+            rules.remove('footnote')
 
-        output = ''
+        output = self.renderer.placeholder()
 
         def manipulate(text):
-            for key in features:
+            for key in rules:
                 pattern = getattr(self.rules, key)
                 m = pattern.match(text)
                 if not m:
@@ -514,7 +519,7 @@ class InlineLexer(object):
                 out = getattr(self, 'output_%s' % key)(m)
                 if out is not None:
                     return m, out
-            return False
+            return False  # pragma: no cover
 
         self.line_started = False
         while text:
@@ -525,7 +530,7 @@ class InlineLexer(object):
                 output += out
                 text = text[len(m.group(0)):]
                 continue
-            if text:
+            if text:  # pragma: no cover
                 raise RuntimeError('Infinite loop at: %s' % text)
 
         return output
@@ -547,14 +552,8 @@ class InlineLexer(object):
             return self.renderer.text(link)
         return self.renderer.autolink(link, False)
 
-    def output_tag(self, m):
-        text = m.group(0)
-        lower_text = text.lower()
-        if lower_text.startswith('<a '):
-            self._in_link = True
-        if lower_text.startswith('</a>'):
-            self._in_link = False
-        return self.renderer.tag(text)
+    def output_inline_html(self, m):
+        return self.renderer.inline_html(m.group(0))
 
     def output_footnote(self, m):
         key = _keyify(m.group(1))
@@ -567,7 +566,7 @@ class InlineLexer(object):
         return self.renderer.footnote_ref(key, self.footnote_index)
 
     def output_link(self, m):
-        return self._process_link(m, m.group(2), m.group(3))
+        return self._process_link(m, m.group(3), m.group(4))
 
     def output_reflink(self, m):
         key = _keyify(m.group(2) or m.group(1))
@@ -627,24 +626,39 @@ class Renderer(object):
     def __init__(self, **kwargs):
         self.options = kwargs
 
+    def placeholder(self):
+        """Returns the default, empty output value for the renderer.
+
+        All renderer methods use the '+=' operator to append to this value.
+        Default is a string so rendering HTML can build up a result string with
+        the rendered Markdown.
+
+        Can be overridden by Renderer subclasses to be types like an empty
+        list, allowing the renderer to create a tree-like structure to
+        represent the document (which can then be reprocessed later into a
+        separate format like docx or pdf).
+        """
+        return ''
+
     def block_code(self, code, lang=None):
         """Rendering block level code. ``pre > code``.
 
         :param code: text content of the code block.
         :param lang: language of the given code.
         """
+        code = code.rstrip('\n')
         if not lang:
             code = escape(code, smart_amp=False)
-            return '<pre><code>%s</code></pre>\n' % code
+            return '<pre><code>%s\n</code></pre>\n' % code
         code = escape(code, quote=True, smart_amp=False)
-        return '<pre><code class="lang-%s">%s</code></pre>\n' % (lang, code)
+        return '<pre><code class="lang-%s">%s\n</code></pre>\n' % (lang, code)
 
     def block_quote(self, text):
         """Rendering <blockquote> with the given text.
 
         :param text: text content of the blockquote.
         """
-        return '<blockquote>%s</blockquote>\n' % text
+        return '<blockquote>%s\n</blockquote>\n' % text.rstrip('\n')
 
     def block_html(self, html):
         """Rendering block level pure html content.
@@ -690,7 +704,7 @@ class Renderer(object):
 
     def paragraph(self, text):
         """Rendering paragraph tags. Like ``<p>``."""
-        return '<p>%s</p>\n' % text
+        return '<p>%s</p>\n' % text.strip(' ')
 
     def table(self, header, body):
         """Rendering table element. Wrap header and body in it.
@@ -753,8 +767,8 @@ class Renderer(object):
     def linebreak(self):
         """Rendering line break like ``<br>``."""
         if self.options.get('use_xhtml'):
-            return '<br />'
-        return '<br>'
+            return '<br />\n'
+        return '<br>\n'
 
     def strikethrough(self, text):
         """Rendering ~~strikethrough~~ text.
@@ -814,17 +828,19 @@ class Renderer(object):
             return '%s />' % html
         return '%s>' % html
 
-    def tag(self, html):
-        """Rendering span level html tag.
+    def inline_html(self, html):
+        """Rendering span level pure html content.
 
-        :param html: html tag snippet.
+        :param html: text content of the html snippet.
         """
-        if self.options.get('skip_html'):
-            return ''
         if self.options.get('escape'):
             return escape(html)
         return html
 
+    def newline(self):
+        """Rendering newline element."""
+        return ''
+
     def footnote_ref(self, key, index):
         """Rendering the ref anchor of a footnote.
 
@@ -905,6 +921,7 @@ class Markdown(object):
 
     def parse(self, text):
         out = self.output(preprocessing(text))
+
         keys = self.block.def_footnotes
 
         # reset block
@@ -923,7 +940,7 @@ class Markdown(object):
             footnotes, key=lambda o: keys.get(o['key']), reverse=True
         )
 
-        body = ''
+        body = self.renderer.placeholder()
         while self.footnotes:
             note = self.footnotes.pop()
             body += self.renderer.footnote_item(
@@ -933,18 +950,6 @@ class Markdown(object):
         out += self.renderer.footnotes(body)
         return out
 
-    def output(self, text, features=None):
-        self.tokens = self.block(text, features)
-        self.tokens.reverse()
-
-        self.inline.setup(self.block.def_links, self.block.def_footnotes)
-
-        out = ''
-        while self.pop():
-            out += self.tok()
-
-        return out
-
     def pop(self):
         if not self.tokens:
             return None
@@ -954,7 +959,18 @@ class Markdown(object):
     def peek(self):
         if self.tokens:
             return self.tokens[-1]
-        return None
+        return None  # pragma: no cover
+
+    def output(self, text, rules=None):
+        self.tokens = self.block(text, rules)
+        self.tokens.reverse()
+
+        self.inline.setup(self.block.def_links, self.block.def_footnotes)
+
+        out = self.renderer.placeholder()
+        while self.pop():
+            out += self.tok()
+        return out
 
     def tok(self):
         t = self.token['type']
@@ -963,7 +979,7 @@ class Markdown(object):
         if t.endswith('_start'):
             t = t[:-6]
 
-        return getattr(self, 'parse_%s' % t)()
+        return getattr(self, 'output_%s' % t)()
 
     def tok_text(self):
         text = self.token['text']
@@ -971,62 +987,65 @@ class Markdown(object):
             text += '\n' + self.pop()['text']
         return self.inline(text)
 
-    def parse_space(self):
-        return ''
+    def output_newline(self):
+        return self.renderer.newline()
 
-    def parse_hrule(self):
+    def output_hrule(self):
         return self.renderer.hrule()
 
-    def parse_heading(self):
+    def output_heading(self):
         return self.renderer.header(
             self.inline(self.token['text']),
             self.token['level'],
             self.token['text'],
         )
 
-    def parse_code(self):
+    def output_code(self):
         return self.renderer.block_code(
             self.token['text'], self.token['lang']
         )
 
-    def parse_table(self):
+    def output_table(self):
         aligns = self.token['align']
-        cell = ''
+        aligns_length = len(aligns)
+        cell = self.renderer.placeholder()
 
         # header part
-        header = ''
+        header = self.renderer.placeholder()
         for i, value in enumerate(self.token['header']):
-            flags = {'header': True, 'align': aligns[i]}
+            align = aligns[i] if i < aligns_length else None
+            flags = {'header': True, 'align': align}
             cell += self.renderer.table_cell(self.inline(value), **flags)
 
         header += self.renderer.table_row(cell)
 
         # body part
-        body = ''
+        body = self.renderer.placeholder()
         for i, row in enumerate(self.token['cells']):
-            cell = ''
+            cell = self.renderer.placeholder()
             for j, value in enumerate(row):
-                flags = {'header': False, 'align': aligns[j]}
+                align = aligns[j] if j < aligns_length else None
+                flags = {'header': False, 'align': align}
                 cell += self.renderer.table_cell(self.inline(value), **flags)
             body += self.renderer.table_row(cell)
 
         return self.renderer.table(header, body)
 
-    def parse_block_quote(self):
-        body = ''
+    def output_block_quote(self):
+        body = self.renderer.placeholder()
         while self.pop()['type'] != 'block_quote_end':
             body += self.tok()
         return self.renderer.block_quote(body)
 
-    def parse_list(self):
+    def output_list(self):
         ordered = self.token['ordered']
-        body = ''
+        body = self.renderer.placeholder()
         while self.pop()['type'] != 'list_end':
             body += self.tok()
         return self.renderer.list(body, ordered)
 
-    def parse_list_item(self):
-        body = ''
+    def output_list_item(self):
+        body = self.renderer.placeholder()
         while self.pop()['type'] != 'list_item_end':
             if self.token['type'] == 'text':
                 body += self.tok_text()
@@ -1035,32 +1054,32 @@ class Markdown(object):
 
         return self.renderer.list_item(body)
 
-    def parse_loose_item(self):
-        body = ''
+    def output_loose_item(self):
+        body = self.renderer.placeholder()
         while self.pop()['type'] != 'list_item_end':
             body += self.tok()
         return self.renderer.list_item(body)
 
-    def parse_footnote(self):
+    def output_footnote(self):
         self.inline._in_footnote = True
-        body = ''
+        body = self.renderer.placeholder()
         key = self.token['key']
         while self.pop()['type'] != 'footnote_end':
             body += self.tok()
         self.footnotes.append({'key': key, 'text': body})
         self.inline._in_footnote = False
-        return ''
+        return self.renderer.placeholder()
 
-    def parse_block_html(self):
+    def output_block_html(self):
         text = self.token['text']
         if self.options.get('parse_html') and not self.token.get('pre'):
             text = self.inline(text)
         return self.renderer.block_html(text)
 
-    def parse_paragraph(self):
+    def output_paragraph(self):
         return self.renderer.paragraph(self.inline(self.token['text']))
 
-    def parse_text(self):
+    def output_text(self):
         return self.renderer.paragraph(self.tok_text())
 
 
@@ -1071,5 +1090,4 @@ def markdown(text, **kwargs):
     :param escape: if set to True, all html tags will be escaped.
     :param use_xhtml: output with xhtml tags.
     """
-    m = Markdown(**kwargs)
-    return m.parse(text)
+    return Markdown(**kwargs)(text)
